from .dandoli import Dandoli
from .mail_dealer import MailDealer
from .microsoft import Microsoft365,SharePoint,PowerApp
from .web_access import WebAccess

__all__ = ['Microsoft365','SharePoint','PowerApp','WebAccess','MailDealer','Dandoli']
